# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from cryptography.hazmat.primitives.ciphers.modes import GCM

from .encryptionalgorithm import EncryptionAlgorithm


class DirectEncryption(EncryptionAlgorithm):
    __module__: str = 'oauthx.algorithms'

    def can_encrypt(self) -> bool:
        return True

    def can_sign(self) -> bool:
        return False

    def get_initialization_vector(self, iv: bytes | None = None, tag: bytes | None = None) -> tuple[GCM, bytes]:
        raise NotImplementedError

    def is_direct(self) -> bool:
        return True

    def supports_aad(self) -> bool:
        return False


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import hashlib
from typing import Any
from typing import Literal

from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES

import aiopki
from aiopki.algorithms import AESEncryption
from aiopki.types import EncryptionResult
from aiopki.types import IAlgorithm
from aiopki.types import Plaintext
from aiopki.utils import b64decode
from aiopki.utils import b64encode
from .basejwk import BaseJWK


class SymmetricEncryptionKey(BaseJWK):
    k: str
    alg: Literal['dir'] # type: ignore
    kty: Literal['oct']
    use: Literal['enc'] | None = 'enc' # type: ignore
    key_ops: list[Literal['encrypt', 'decrypt']] = ['encrypt', 'decrypt'] # type: ignore

    @property
    def public(self) -> None:
        return None

    @property
    def thumbprint(self) -> str:
        message = self.model_dump_json(include={'k', 'kty'})
        return b64encode(hashlib.sha256(str.encode(message)).digest(), encoder=bytes.decode)

    def can_use(self, algorithm: IAlgorithm) -> bool:
        return all([
            isinstance(algorithm, AESEncryption),
            len(self._key) == algorithm.length(),
        ])

    def default_algorithm(self) -> IAlgorithm:
        assert self.use is not None
        algorithm: IAlgorithm = aiopki.algorithms.get('notimplemented')
        if self.use == 'enc':
            algorithm = aiopki.algorithms.get('A256GCM')
        return algorithm

    def encrypt_sync(
        self,
        pt: bytes | Plaintext,
        algorithm: IAlgorithm | None = None
    ) -> EncryptionResult:
        if isinstance(pt, bytes):
            pt = Plaintext(pt=pt)
        algorithm = algorithm or self.default_algorithm()
        mode, iv = algorithm.get_initialization_vector()
        cipher = Cipher(AES(self._key), mode)
        encrypter = cipher.encryptor()
        if algorithm.supports_aad():
            encrypter.authenticate_additional_data(pt.aad or b'')
        ct = encrypter.update(bytes(pt)) + encrypter.finalize()
        return EncryptionResult.model_validate({
            'aad': pt.aad or b'',
            'alg': algorithm.name,
            'kid': self.kid,
            'ct': ct,
            'iv': iv,
            'tag': encrypter.tag
        })

    def model_post_init(self, *args: Any, **kwargs: Any) -> None:
        self._key = b64decode(self.k)

    def to_bytes(self) -> bytes:
        return self._key

    async def decrypt(
        self,
        ct: EncryptionResult,
        algorithm: IAlgorithm | None = None
    ) -> Plaintext:
        assert ct.root.tag is not None
        algorithm = algorithm or self.default_algorithm()
        mode, _ = algorithm.get_initialization_vector(ct.root.iv, ct.root.tag)
        cipher = Cipher(AES(self._key), mode)
        decrypter = cipher.decryptor()
        if algorithm.supports_aad():
            decrypter.authenticate_additional_data(ct.aad)
        return Plaintext.model_validate({
            'pt': decrypter.update(bytes(ct)) + decrypter.finalize(),
            'aad': ct.aad
        })

    async def encrypt(
        self,
        pt: bytes | Plaintext,
        algorithm: IAlgorithm | None = None
    ) -> EncryptionResult:
        return self.encrypt_sync(pt, algorithm)

    async def sign( # type: ignore
        self,
        message: bytes,
        algorithm: IAlgorithm,
    ) -> bytes:
        raise NotImplementedError

    async def verify( # type: ignore
        self,
        signature: bytes,
        message: bytes,
        algorithm: IAlgorithm,
        using: str | None = None
    ) -> bool:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic


JOSE_CTY: set[str] = {
    'jose',
    'jose+json',
    'application/jose',
    'application/jose+json',
}


class Payload(pydantic.BaseModel):
    cty: str | None = None
    value: bytes

    @property
    def typ(self) -> None:
        return None

    def is_compact(self) -> bool:
        return self.cty in {'jose', 'application/json'}

    def is_jose(self) -> bool:
        return self.cty in JOSE_CTY

    def __bytes__(self) -> bytes:
        return self.value


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Literal

from .symmetricencryptionkey import SymmetricEncryptionKey


class SymmetricWrappingKey(SymmetricEncryptionKey):
    alg: Literal['A128CBC-HS256', 'A192CBC-HS384', 'A256CBC-HS512', 'A128GCM', 'A192GCM', 'A256GCM'] # type: ignore
    key_ops: list[Literal['wrapKey', 'unwrapKey']] = ['wrapKey', 'unwrapKey'] # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os
from typing import Any

import pytest
from pydantic import TypeAdapter

from aiopki import CryptoKeyType
from aiopki.utils import b64encode


CryptoKeyAdapter = TypeAdapter(CryptoKeyType)


@pytest.mark.parametrize("v", [
    {'kty': 'oct', 'alg': 'HS256', 'use': 'sig', 'k': b64encode(os.urandom(16))},
    'file:///pki/enc.key?kty=RSA&use=enc&alg=RSA-OAEP-256',
])
def test_parse_specification(v: Any):
    CryptoKeyAdapter.validate_python(v)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Callable
from typing import TypeVar

import pydantic


T = TypeVar('T')


class SubjectID(pydantic.BaseModel):
    iss: str
    sub: str

    @property
    def maskable(self) -> bytes:
        return str.encode(f'subject:{self.iss}/{self.sub}', 'utf-8')

    def mask(self, masker: Callable[[bytes], T]) -> T:
        return masker(self.maskable)

    def __hash__(self) -> int:
        return hash((self.iss, self.sub))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
from typing import Any
from typing import Literal

import pydantic

from aiopki.types import Base64
from aiopki.utils import b64encode
from aiopki.utils import b64decode_json
from .jwa import JWA
from .jwk import JWK


EncryptionAlgorithm = Literal[
    JWA.a128gcm,
    JWA.a192gcm,
    JWA.a256gcm,
]


class JWEHeader(pydantic.BaseModel):
    alg: JWA
    cty: str | None = None
    crit: list[str] = []
    jku: str | None = None
    jwk: JWK | None = None
    kid: str | None = None
    typ: str | None = None
    x5c: str | None = None
    x5t: str | None = None
    x5t_sha256: str | None = pydantic.Field(default=None, alias='x5t#S256')
    x5u: str | None = None
    claims: dict[str, Any] = {}

    enc: Literal[EncryptionAlgorithm]
    tag: Base64 | None = None
    iv: Base64 | None = None
    zip: str | None = None

    @pydantic.model_validator(mode='before')
    def preprocess(cls, values: dict[str, Any] | str) -> dict[str, Any]:
        if isinstance(values, str):
            result = b64decode_json(values)
            if not isinstance(result, dict):
                raise ValueError("The JWE Protected Header must be a JSON object.")
            values = result
        return {
            **{
                name: values[name] for name in cls.model_fields.keys()
                if name in values and name != 'claims'
            },
            **(values.get('claims') or {})
        }

    def encode(self, **kwargs: Any) -> bytes:
        kwargs.setdefault('exclude_none', True)
        kwargs.setdefault('exclude_defaults', True)
        kwargs.setdefault('exclude_unset', True)
        exclude = kwargs.setdefault('exclude', set())
        exclude.add('claims')
        claims = self.model_dump(**kwargs)
        claims.update(self.claims)
        return b64encode(json.dumps(claims))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from cryptography.hazmat.primitives.hashes import Hash
from cryptography.hazmat.primitives.hashes import HashAlgorithm
from cryptography.hazmat.primitives.hashes import SHA1
from cryptography.hazmat.primitives.hashes import SHA256
from cryptography.hazmat.primitives.hashes import SHA384
from cryptography.hazmat.primitives.hashes import SHA512
from cryptography.hazmat.primitives.asymmetric.padding import AsymmetricPadding
from cryptography.hazmat.primitives.asymmetric.padding import MGF1
from cryptography.hazmat.primitives.asymmetric.padding import OAEP
from cryptography.hazmat.primitives.asymmetric.padding import PSS
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15

from ..types import IAlgorithm


DIGEST_ALGORITHMS: dict[str, type[HashAlgorithm]] = {
    'sha1'  : SHA1,
    'sha256': SHA256,
    'sha384': SHA384,
    'sha512': SHA512,
}

PADDING_ALGORITHMS: dict[str, type[OAEP|PKCS1v15|PSS]] = {
    'EME-OAEP': OAEP,
    'RSASSA-PKCS1-v1_5': PKCS1v15
}


class RSAAlgorithm(IAlgorithm):
    __module__: str = 'aiopki.algorithms'
    name: str

    def __init__(self, digalg: str, padding: str):
        self._digalg = DIGEST_ALGORITHMS[digalg]
        self._padding = PADDING_ALGORITHMS[padding]
        self._padding_name = padding

    def can_encrypt(self) -> bool:
        return False

    def can_sign(self) -> bool:
        return False

    def digest(self, value: bytes) -> bytes:
        h = Hash(self._digalg())
        h.update(value)
        return h.finalize()

    def get_curve_name(self) -> str | None:
        return None

    def get_digest_algorithm(self) -> HashAlgorithm:
        return self._digalg()
    
    def get_padding(self) -> AsymmetricPadding:
        params = self.get_padding_params(self._padding_name)
        if self._padding == PSS:
            raise NotImplementedError
        return self._padding(**params) # type: ignore

    def get_padding_params(self, padding: str) -> dict[str, Any]:
        params: dict[str, Any] = {}
        match padding:
            case 'EME-OAEP':
                params.update({
                    'mgf': MGF1(self.get_digest_algorithm()),
                    'algorithm': self.get_digest_algorithm(),
                    'label': None
                })
            case _:
                pass
        return params


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import os

from aiopki.tests.compat.jwe import *
from aiopki.utils import b64encode


@pytest.fixture(scope='module')
def alg() -> str:
    return 'dir'


@pytest.fixture(scope='module')
def enc() -> str:
    return 'A192GCM'


@pytest.fixture(scope='module')
def encrypter(alg: str) -> JSONWebKey:
    return JSONWebKey.model_validate({
        'kty': 'oct',
        'use': 'enc',
        'alg': alg,
        'k': b64encode(os.urandom(24))
    })


@pytest.fixture(scope='module')
def decrypter(encrypter: JSONWebKey) -> JSONWebKey:
    return encrypter


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

import pydantic

import aiopki
from aiopki.algorithms import DirectEncryption
from aiopki.lib import JSONWebKey
from aiopki.types import Base64
from aiopki.types import IAlgorithm
from aiopki.types import IDecrypter
from aiopki.types import EncryptionResult
from aiopki.types import JSONObject
from aiopki.types import Plaintext

from .jweheader import JWEHeader


class Recipient(pydantic.BaseModel):
    header: JSONObject = pydantic.Field(
        default_factory=JSONObject
    )

    encrypted_key: Base64 | None = pydantic.Field(
        default=None
    )

    _claims: JWEHeader = pydantic.PrivateAttr(default_factory=dict)

    @property
    def alg(self) -> IAlgorithm:
        return aiopki.get(self._claims.alg.value)

    @property
    def enc(self) -> IAlgorithm:
        return aiopki.get(self._claims.enc.value)

    @property
    def iv(self) -> bytes | None:
        return self._claims.iv

    @property
    def kid(self) -> str | None:
        return self._claims.kid

    @property
    def tag(self) -> bytes | None:
        return self._claims.tag

    def is_direct(self) -> bool:
        return self._claims.alg.name == 'dir'

    def update_header(self, claims: dict[str, Any]):
        self._claims = JWEHeader.model_validate({**claims, **self.header})
        if self.alg.name == 'dir' and self.encrypted_key:
            raise ValueError("Direct encryption should not provide an encryption key.")
        elif self.alg.name != 'dir' and not self.encrypted_key:
            raise ValueError("Wrapped encryption must provide an encryption key.")

    def can_decrypt(self, decrypter: IDecrypter) -> bool:
        return all([
            decrypter.can_use(self.alg if not self.is_direct() else self.enc)
        ])

    async def decrypt(self, decrypter: IDecrypter, ct: EncryptionResult) -> Plaintext:
        if not isinstance(self.alg, DirectEncryption):
            decrypter = await self.unwrap(decrypter)
        return await decrypter.decrypt(
            ct=ct,
            algorithm=self.enc
        )

    async def unwrap(self, decrypter: IDecrypter) -> IDecrypter:
        return await self.enc.unwrap(
            algorithm=self.alg,
            decrypter=decrypter,
            ct=EncryptionResult.model_validate({
                'alg': self.enc.name,
                'ct': self.encrypted_key
            }),
            factory=JSONWebKey.model_validate
        )